/*
 ************************************************************************
 *
 * File 'modetest1.c'.
 *
 * CanvasPlus widget test mode implementation.
 *
 * Written by Jorge Lima in Nov 23 14:44:03 WET 2004
 *
 */
#include <tcl.h>
#include <math.h>

#include "tkCanvasPlusTags.h"
#include "tkCanvasPlus.h"
#include "tkCanvasPlusDrawing.h"
#include "tkCanvasPlusModes.h"


typedef struct MStatus {
  int in_use;
  CanvasPlus *canvas;
  CP_Item *item;
} MStatus;

typedef enum ItemType {ARC_ITEM, RECT_ITEM} ItemType;

typedef struct ArcItem {
  CP_Item item;
  CP_Area limits;
  // only needed to speedup calculations
  CP_Point center;
  CP_Point haxis;
  char *color;
} ArcItem;

typedef struct RectItem {
  CP_Item item;
  CP_Area limits;
  char *color;
} RectItem;


static MStatus _status = { 0, NULL };
 
/* FREE PROC */
static void freeProc(ClientData cd)
{
  _status.in_use = 0;
}

static ClientData allocProc(CanvasPlus *canvas)
{
  if (!_status.in_use) {
    _status.in_use = 1;
    _status.canvas = canvas;
    _status.item = NULL;
    return (ClientData)&_status;
  }
  else {
    return NULL;
  }
}

static int initProc(CP_Mode *mode) 
{
  if (_status.in_use) {
    unsigned long mask;
    XGCValues values;

    mask = GCFunction|GCForeground|GCBackground|
      GCLineWidth|GCLineStyle|GCCapStyle|GCJoinStyle;
    values.function = GXcopy;
    values.foreground = 0x0F0F0FFF;
    values.background = 0xFFFFFFFF;
    values.line_width = 0;
    values.line_style = LineSolid;
    values.cap_style = CapButt;
    values.join_style = JoinMiter;

    CPCreatePen(mode, &values, mask); //pen 0
    values.foreground = 0x0F0FFF0F;
    values.background = 0xFFFFFFFF;
    CPCreatePen(mode, &values, mask); //pen 1
    return 0;
  }
  return 1;
}

static void drawProc(ClientData cd, CP_DrawOps *ops, CP_Area *a)
{
  MStatus *status = (MStatus *)cd;
  CP_Item *item = status->item;
 
  while (item != NULL) {
    ItemType type = (ItemType)item->clientData;
    switch (type) {
      case ARC_ITEM: {
	ArcItem *arc = (ArcItem *)item;
	ops->selectPen(ops->clientData, 0);
	//ops->setPenColor(ops->clientData, "red", "orange");
	ops->setPenWidth(ops->clientData, 1.5);
	ops->drawElipse(ops->clientData, &arc->limits, CP_DRAW_OUTLINE);
	break;
      }
      case RECT_ITEM: {
	RectItem *rect = (RectItem *)item;
	ops->selectPen(ops->clientData, 1);
	//	ops->setPenColor(ops->clientData, "yellow", "blue");
	ops->drawRectangle(ops->clientData, &rect->limits, CP_DRAW_FILLED);
	ops->selectPen(ops->clientData, 0);
	ops->setPenWidth(ops->clientData, 1.5);
	ops->drawRectangle(ops->clientData, &rect->limits, CP_DRAW_OUTLINE);
	break;
      }
    }
    item = item->next;
  }
}

inline static double norm(double x1, double y1, double x2, double y2)
{
  double dx = abs(x2 - x1); 
  double dy = abs(y2 - y1);
  if (dx<dy)
    return dy * sqrt(pow(dx/dy,2) + 1);
  else 
    return dx * sqrt(pow(dy/dx,2) + 1);
}

static double distanceToPoint(CP_Item *item, CP_Point *pt)
{
  ItemType type = (ItemType)item->clientData;
  switch (type) {
  case ARC_ITEM: {
    ArcItem *arc = (ArcItem *)item;
    double dx = pt->x - arc->center.x;
    double dy = pt->y - arc->center.y;
    double tmp = pow(dx/arc->haxis.x,2) + pow(dy/arc->haxis.y,2);
    if (tmp<=1)
      return 0;
    else
      return sqrt(tmp-1);
    break;
  }
  case RECT_ITEM: {
    RectItem *rect = (RectItem *)item;
    CP_Area *area = &rect->limits;
    if (pt->x < area->x1) {
      if (pt->y < area->y1)
	return norm(pt->x, pt->y, area->x1, area->y1);
      else if (pt->y > area->y2)
	return norm(pt->x, pt->y, area->x1, area->y2);
      else 
	return area->x1 - pt->x;
    }  
    else if (pt->x > area->x2) {
      if (pt->y < area->y1)
	return norm(pt->x, pt->y, area->x2, area->y1);
      else if (pt->y > area->y2)
	return norm(pt->x, pt->y, area->x2, area->y2);
      else 
	return pt->x - area->x2;
    }
    else {
      if (pt->y < area->y1)
	return area->y1 - pt->y;
      else if (pt->y > area->y2)
	return pt->y - area->y2;
      else
	return 0.0;
    }
    break;
  }
  }
  return -1.0;
}

static CP_Item *closestProc(ClientData cd, CP_Point *pt, double dist)
{
  MStatus *status = (MStatus *)cd;
  CP_Item *item = status->item;
  CP_Item *best = NULL;
  while (item != NULL) {
    if (distanceToPoint(item, pt)<=dist)
      best = item;
    item = item->next;
  }
  return best;
}
 
/* PROCS TO REGISTER */
static int putsProc( ClientData clientData,
		     Tcl_Interp *interp,
		     int objc,
		     Tcl_Obj *CONST objv[])
{
  int i;
  for(i = 1; i < objc; i++)
    printf ("%s ",Tcl_GetStringFromObj(objv[i], NULL));
  printf("\n");
  return TCL_OK;
}

static int moveProc( ClientData clientData,
		     Tcl_Interp *interp,
		     int objc,
		     Tcl_Obj *CONST objv[])
{
  MStatus *status = (MStatus *)clientData;
  CP_Item *item = status->canvas->currentItem;
  double x1, y1, x2, y2;

  if (objc != 5) {
    Tcl_AppendResult(interp, "usage: @move x1 y1 x2 y2");
    return TCL_ERROR;
  }
  if (Tcl_GetDoubleFromObj(interp, objv[1], &x1)!=TCL_OK)
    return TCL_ERROR;
  if (Tcl_GetDoubleFromObj(interp, objv[2], &y1)!=TCL_OK)
    return TCL_ERROR;
  if (Tcl_GetDoubleFromObj(interp, objv[3], &x2)!=TCL_OK)
    return TCL_ERROR;
  if (Tcl_GetDoubleFromObj(interp, objv[4], &y2)!=TCL_OK)
    return TCL_ERROR;

  if (item != NULL) {
    ItemType type = (ItemType)item->clientData;
    switch (type) {
    case ARC_ITEM:
      {
	ArcItem *arc = (ArcItem *)item;
	double dx = x2-x1, dy = y2-y1;
	arc->limits.x1 += dx, arc->limits.x2 += dx;
	arc->limits.y1 += dy, arc->limits.y2 += dy;
	arc->center.x += dx, arc->center.y += dy;
	CPExpandPhysicalDimensions(status->canvas, &arc->limits);
      }
      break;
    case RECT_ITEM:
      {
	RectItem *rect = (RectItem *)item;
	double dx = x2-x1, dy = y2-y1;
	rect->limits.x1 += dx, rect->limits.x2 += dx;
	rect->limits.y1 += dy, rect->limits.y2 += dy;
	CPExpandPhysicalDimensions(status->canvas, &rect->limits);
      }
      break;
    }
    item = item->next;
  }
  return TCL_OK;
}

static int rectProc( ClientData clientData,
		     Tcl_Interp *interp,
		     int objc,
		     Tcl_Obj *CONST objv[])
{
  MStatus *status = (MStatus *)clientData;
  CP_Item *item;
  RectItem *rect;
  unsigned size;

  if (objc != 5) {
    Tcl_AppendResult(interp, "usage: @rect x1 y1 x2 y2");
    return TCL_ERROR;
  }

  // Create a new Item. Reserve private data
  size = sizeof(RectItem)-sizeof(CP_Item);
  item = CPCreateItem(status->canvas, size);
  if (item==NULL) return TCL_ERROR;

  // Initialize items client data
  item->clientData = (ClientData)RECT_ITEM;

  // initialize rectangle
  rect = (RectItem *)item;
  if (Tcl_GetDoubleFromObj(interp, objv[1], &rect->limits.x1)!=TCL_OK)
    goto error;
  if (Tcl_GetDoubleFromObj(interp, objv[2], &rect->limits.y1)!=TCL_OK)
    goto error;
  if (Tcl_GetDoubleFromObj(interp, objv[3], &rect->limits.x2)!=TCL_OK)
    goto error;
  if (Tcl_GetDoubleFromObj(interp, objv[4], &rect->limits.y2)!=TCL_OK)
    goto error;

  // Make sure the canvas is expanded to contain at least this item
  CPExpandPhysicalDimensions(status->canvas, &rect->limits);
  CPAddBindTag(item, CPGetTagUid("rectangle"));

  // Test this is the first item.
  if (status->item == NULL) status->item = item;
  Tcl_SetObjResult(interp, Tcl_NewIntObj(item->id));
  return TCL_OK;
 error:
  CPDeleteItem(status->canvas, item);
  return TCL_ERROR;
}

static int arcProc( ClientData clientData,
		    Tcl_Interp *interp,
		    int objc,
		    Tcl_Obj *CONST objv[])
{
  MStatus *status = (MStatus *)clientData;
  CP_Item *item;
  ArcItem *arc;
  unsigned size;

  if (objc != 5) {
    Tcl_AppendResult(interp, "usage: @elipse x1 y1 x2 y2");
    return TCL_ERROR;
  }

  // Create a new Item. Reserve private data
  size = sizeof(ArcItem)-sizeof(CP_Item);
  item = CPCreateItem(status->canvas, size);
  if (item==NULL) return TCL_ERROR;

  // Initialize items client data
  item->clientData = (ClientData)ARC_ITEM;

  // initialize rectangle
  arc = (ArcItem *)item;
  if (Tcl_GetDoubleFromObj(interp, objv[1], &arc->limits.x1)!=TCL_OK)
    goto error;
  if (Tcl_GetDoubleFromObj(interp, objv[2], &arc->limits.y1)!=TCL_OK)
    goto error;
  if (Tcl_GetDoubleFromObj(interp, objv[3], &arc->limits.x2)!=TCL_OK)
    goto error;
  if (Tcl_GetDoubleFromObj(interp, objv[4], &arc->limits.y2)!=TCL_OK)
    goto error;

  arc->center.x = (arc->limits.x2 + arc->limits.x1) / 2;
  arc->center.y = (arc->limits.y2 + arc->limits.y1) / 2;
  arc->haxis.x = abs(arc->limits.x2 - arc->center.x);
  arc->haxis.y = abs(arc->limits.y2 - arc->center.y);

  // Make sure the canvas is expanded to contain at least this item
  CPExpandPhysicalDimensions(status->canvas, &arc->limits);
  CPAddBindTag(item, CPGetTagUid("elipse"));

  // Test this is the first item.
  if (status->item == NULL) status->item = item;
  Tcl_SetObjResult(interp, Tcl_NewIntObj(item->id));

  return TCL_OK;
 error:
  CPDeleteItem(status->canvas, item);
  return TCL_ERROR;
}

/* REGISTER INFO */
static CP_RegisterInfo _info[] = {
  CP_ENTRY("@puts", putsProc),
  CP_ENTRY("@rect", rectProc),
  CP_ENTRY("@elipse", arcProc),
  CP_ENTRY("@move", moveProc),
  CP_LAST_ENTRY
};

static CP_ModeOps _operations = {
  allocProc,
  initProc,
  freeProc,
  drawProc,
  closestProc
};

/*
 * INIT 
 */

int Modetest_Init(Tcl_Interp *interp)
{
  if ( CPDefineMode("modetest1", _info, &_operations)!=TCL_OK )
    return TCL_ERROR;

  return TCL_OK;
}


